import functools
import os

from dagster import InputDefinition, List, Output, OutputDefinition, Path, SolidDefinition, check

from .configs import define_spark_config
from .types import SparkSolidError
from .utils import parse_spark_config, run_spark_subprocess


def create_spark_shell_cmd(solid_config, main_class):
    # Extract parameters from config
    (master_url, deploy_mode, application_jar, spark_conf, application_arguments, spark_home) = [
        solid_config.get(k)
        for k in (
            'master_url',
            'deploy_mode',
            'application_jar',
            'spark_conf',
            'application_arguments',
            'spark_home',
        )
    ]

    # Let the user use env vars in the jar path
    application_jar = os.path.expandvars(application_jar)

    if not os.path.exists(application_jar):
        raise SparkSolidError(
            (
                'Application jar {} does not exist. A valid jar must be '
                'built before running this solid.'.format(application_jar)
            )
        )

    spark_home = spark_home if spark_home else os.environ.get('SPARK_HOME')

    if spark_home is None:
        raise SparkSolidError(
            (
                'No spark home set. You must either pass spark_home in config or '
                'set $SPARK_HOME in your environment (got None).'
            )
        )

    deploy_mode = ['--deploy-mode', '{}'.format(deploy_mode)] if deploy_mode else []

    spark_shell_cmd = (
        ['{}/bin/spark-submit'.format(spark_home), '--class', main_class, '--master', master_url]
        + deploy_mode
        + parse_spark_config(spark_conf)
        + [application_jar]
        + ([application_arguments] if application_arguments else [])
    )
    return spark_shell_cmd


def step_metadata_fn(environment_config, solid_name, main_class):
    return {
        'spark_submit_command': ' '.join(
            create_spark_shell_cmd(environment_config.solids[solid_name].config, main_class)
        )
    }


class SparkSolidDefinition(SolidDefinition):
    '''This solid is a generic representation of a parameterized Spark job.

    Parameters:
    name (str): The name of this Spark solid
    main_class (str): The entry point for your application (e.g. org.apache.spark.examples.SparkPi)

    '''

    def __init__(self, name, main_class, description=None):
        name = check.str_param(name, 'name')
        main_class = check.str_param(main_class, 'main_class')
        description = check.opt_str_param(
            description,
            'description',
            'This solid is a generic representation of a parameterized Spark job.',
        )

        def _spark_compute_fn(context, _):
            '''Define Spark execution.

            This function defines how we'll execute the Spark job and invokes spark-submit.
            '''

            spark_shell_cmd = create_spark_shell_cmd(context.solid_config, main_class)

            context.log.info("Running spark-submit: " + ' '.join(spark_shell_cmd))
            retcode = run_spark_subprocess(spark_shell_cmd, context.log)

            if retcode != 0:
                raise SparkSolidError('Spark job failed. Please consult your logs.')

            yield Output(context.solid_config.get('spark_outputs'), 'paths')

        super(SparkSolidDefinition, self).__init__(
            name=name,
            description=description,
            input_defs=[InputDefinition('spark_inputs', List[Path])],
            output_defs=[OutputDefinition(dagster_type=List[Path], name='paths')],
            compute_fn=_spark_compute_fn,
            config_field=define_spark_config(),
            metadata={'kind': 'spark', 'main_class': main_class},
            step_metadata_fn=functools.partial(
                step_metadata_fn, solid_name=name, main_class=main_class
            ),
        )



import itertools
import re
import subprocess
from threading import Thread

from six.moves import queue


def run_spark_subprocess(cmd, logger):
    """See https://bit.ly/2OpksJC for source of the subprocess stdout/stderr capture pattern in this
    function.
    """

    # Spark sometimes logs in log4j format. In those cases, we detect and parse.
    # Example log line from Spark that this is intended to match:
    # 2019-03-27 16:00:19 INFO  ContextHandler:781 - Started o.s.j.s.ServletContextHandler...
    log4j_regex = r'^(\d{4}\-\d{2}\-\d{2} \d{2}:\d{2}:\d{2}) ([A-Z]{3,5})(.*?)$'

    def reader(pipe, pipe_name, p, msg_queue):
        try:
            with pipe:
                while p.poll() is None:
                    for line in pipe.readlines():
                        match = re.match(log4j_regex, line)
                        if match:
                            line = match.groups()[2]
                        msg_queue.put((pipe_name, line))
        finally:
            # Use None as sentinel for done state, detected by iter() below
            msg_queue.put(None)

    p = subprocess.Popen(
        ' '.join(cmd),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        bufsize=0,
        universal_newlines=True,
        shell=True,
    )
    q = queue.Queue()
    Thread(target=reader, args=[p.stdout, 'stdout', p, q]).start()
    Thread(target=reader, args=[p.stderr, 'stderr', p, q]).start()
    for _ in range(2):  # There will be two None sentinels, one for each stream
        for pipe_name, line in iter(q.get, None):
            if pipe_name == 'stdout':
                logger.info(line)
            elif pipe_name == 'stderr':
                logger.error(line)

    p.wait()
    return p.returncode


def flatten_dict(d):
    def _flatten_dict(d, result, key_path=None):
        '''Iterates an arbitrarily nested dictionary and yield dot-notation key:value tuples.

        {'foo': {'bar': 3, 'baz': 1}, {'other': {'key': 1}} =>
            [('foo.bar', 3), ('foo.baz', 1), ('other.key', 1)]

        '''
        for k, v in d.items():
            new_key_path = (key_path or []) + [k]
            if isinstance(v, dict):
                _flatten_dict(v, result, new_key_path)
            else:
                result.append(('.'.join(new_key_path), v))

    result = []
    _flatten_dict(d, result)
    return result


def parse_spark_config(spark_conf):
    '''For each key-value pair in spark conf, we need to pass to CLI in format:

        --conf "key=value"
    '''

    spark_conf_list = flatten_dict(spark_conf)
    return list(
        itertools.chain.from_iterable([('--conf', '{}={}'.format(*c)) for c in spark_conf_list])
    )



import os
import uuid

import pytest
import yaml

from dagster import PipelineDefinition, execute_pipeline
from dagster.core.execution.api import create_execution_plan
from dagster.utils import script_relative_path

from dagster_spark import SparkSolidDefinition, SparkSolidError

CONFIG_FILE = '''
solids:
  spark_solid:
    inputs:
      spark_inputs: []
    config:
      spark_home: /your/spark_home
      spark_outputs: ["/tmp/dagster/events/data"]
      application_jar: "{path}"
      deploy_mode: "client"
      application_arguments: "--local-path /tmp/dagster/events/data --date 2019-01-01"
      master_url: "local[*]"
      spark_conf:
        spark:
          app:
            name: "test_app"
'''


def test_step_metadata():
    spark_solid = SparkSolidDefinition('spark_solid', main_class='something')
    pipeline = PipelineDefinition(solid_defs=[spark_solid])
    environment_dict = yaml.load(CONFIG_FILE.format(path=script_relative_path('fake.jar')))
    execution_plan = create_execution_plan(pipeline, environment_dict)

    step = execution_plan.get_step_by_key('spark_solid.compute')
    assert step.metadata == {
        'spark_submit_command': (
            '/your/spark_home/bin/spark-submit --class something '
            '--master local[*] --deploy-mode client --conf spark.app.name=test_app '
            + script_relative_path('fake.jar')
            + ' --local-path /tmp/dagster/events/data '
            '--date 2019-01-01'
        )
    }


def test_jar_not_found():
    spark_solid = SparkSolidDefinition('spark_solid', main_class='something')
    pipeline = PipelineDefinition(solid_defs=[spark_solid])
    # guid guaranteed to not exist
    environment_dict = yaml.load(CONFIG_FILE.format(path=str(uuid.uuid4())))
    with pytest.raises(
        SparkSolidError,
        match='does not exist. A valid jar must be built before running this solid.',
    ):
        execute_pipeline(pipeline, environment_dict)


NO_SPARK_HOME_CONFIG_FILE = '''
solids:
  spark_solid:
    inputs:
      spark_inputs: []
    config:
      spark_outputs: ["/tmp/dagster/events/data"]
      application_jar: "{path}"
      deploy_mode: "client"
      application_arguments: "--local-path /tmp/dagster/events/data --date 2019-01-01"
      master_url: "local[*]"
      spark_conf:
        spark:
          app:
            name: "test_app"
'''


def test_no_spark_home():
    if 'SPARK_HOME' in os.environ:
        del os.environ['SPARK_HOME']

    spark_solid = SparkSolidDefinition('spark_solid', main_class='something')
    pipeline = PipelineDefinition(solid_defs=[spark_solid])
    environment_dict = yaml.load(NO_SPARK_HOME_CONFIG_FILE.format(path=script_relative_path('.')))

    with pytest.raises(SparkSolidError) as exc_info:
        execute_pipeline(pipeline, environment_dict)

    assert str(exc_info.value) == (
        'No spark home set. You must either pass spark_home in config or set '
        '$SPARK_HOME in your environment (got None).'
    )



'''NOTE: THIS FILE IS AUTO-GENERATED. DO NOT EDIT

@generated

Produced via:
parse_spark_configs.py \
	--output-file \
	../libraries/dagster-spark/dagster_spark/configs_spark.py \

'''


from dagster import Field, PermissiveDict, String


# pylint: disable=line-too-long
def spark_config():
    return Field(
        PermissiveDict(
            fields={
                'spark': Field(
                    PermissiveDict(
                        fields={
                            'app': Field(
                                PermissiveDict(
                                    fields={
                                        'name': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        )
                                    }
                                )
                            ),
                            'driver': Field(
                                PermissiveDict(
                                    fields={
                                        'cores': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'maxResultSize': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'memory': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'memoryOverhead': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'supervise': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'extraClassPath': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'extraJavaOptions': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'extraLibraryPath': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'userClassPathFirst': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'blockManager': Field(
                                            PermissiveDict(
                                                fields={
                                                    'port': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        ),
                                        'bindAddress': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'host': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'port': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                    }
                                )
                            ),
                            'executor': Field(
                                PermissiveDict(
                                    fields={
                                        'memory': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'pyspark': Field(
                                            PermissiveDict(
                                                fields={
                                                    'memory': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        ),
                                        'memoryOverhead': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'extraClassPath': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'extraJavaOptions': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'extraLibraryPath': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'logs': Field(
                                            PermissiveDict(
                                                fields={
                                                    'rolling': Field(
                                                        PermissiveDict(
                                                            fields={
                                                                'maxRetainedFiles': Field(
                                                                    String,
                                                                    description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                                    is_optional=True,
                                                                ),
                                                                'enableCompression': Field(
                                                                    String,
                                                                    description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                                    is_optional=True,
                                                                ),
                                                                'maxSize': Field(
                                                                    String,
                                                                    description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                                    is_optional=True,
                                                                ),
                                                                'strategy': Field(
                                                                    String,
                                                                    description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                                    is_optional=True,
                                                                ),
                                                                'time': Field(
                                                                    PermissiveDict(
                                                                        fields={
                                                                            'interval': Field(
                                                                                String,
                                                                                description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                                                is_optional=True,
                                                                            )
                                                                        }
                                                                    )
                                                                ),
                                                            }
                                                        )
                                                    )
                                                }
                                            )
                                        ),
                                        'userClassPathFirst': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'cores': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'heartbeatInterval': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                    }
                                )
                            ),
                            'extraListeners': Field(
                                String,
                                description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                is_optional=True,
                            ),
                            'local': Field(
                                PermissiveDict(
                                    fields={
                                        'dir': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        )
                                    }
                                )
                            ),
                            'logConf': Field(
                                String,
                                description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                is_optional=True,
                            ),
                            'master': Field(
                                String,
                                description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                is_optional=True,
                            ),
                            'submit': Field(
                                PermissiveDict(
                                    fields={
                                        'deployMode': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'pyFiles': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                    }
                                )
                            ),
                            'log': Field(
                                PermissiveDict(
                                    fields={
                                        'callerContext': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        )
                                    }
                                )
                            ),
                            'redaction': Field(
                                PermissiveDict(
                                    fields={
                                        'regex': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        )
                                    }
                                )
                            ),
                            'python': Field(
                                PermissiveDict(
                                    fields={
                                        'profile': Field(
                                            PermissiveDict(
                                                fields={
                                                    'root': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'dump': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                }
                                            )
                                        ),
                                        'worker': Field(
                                            PermissiveDict(
                                                fields={
                                                    'memory': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'reuse': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                }
                                            )
                                        ),
                                    }
                                )
                            ),
                            'files': Field(
                                PermissiveDict(
                                    fields={
                                        'root': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'fetchTimeout': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'useFetchCache': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'overwrite': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'maxPartitionBytes': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'openCostInBytes': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                    }
                                )
                            ),
                            'jars': Field(
                                PermissiveDict(
                                    fields={
                                        'root': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'packages': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'excludes': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'ivy': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'ivySettings': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'repositories': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                    }
                                )
                            ),
                            'pyspark': Field(
                                PermissiveDict(
                                    fields={
                                        'driver': Field(
                                            PermissiveDict(
                                                fields={
                                                    'python': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        ),
                                        'python': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                    }
                                )
                            ),
                            'reducer': Field(
                                PermissiveDict(
                                    fields={
                                        'maxSizeInFlight': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'maxReqsInFlight': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'maxBlocksInFlightPerAddress': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                    }
                                )
                            ),
                            'maxRemoteBlockSizeFetchToMem': Field(
                                String,
                                description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                is_optional=True,
                            ),
                            'shuffle': Field(
                                PermissiveDict(
                                    fields={
                                        'compress': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'file': Field(
                                            PermissiveDict(
                                                fields={
                                                    'buffer': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        ),
                                        'io': Field(
                                            PermissiveDict(
                                                fields={
                                                    'maxRetries': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'numConnectionsPerPeer': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'preferDirectBufs': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'retryWait': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                }
                                            )
                                        ),
                                        'service': Field(
                                            PermissiveDict(
                                                fields={
                                                    'enabled': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'port': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'index': Field(
                                                        PermissiveDict(
                                                            fields={
                                                                'cache': Field(
                                                                    PermissiveDict(
                                                                        fields={
                                                                            'size': Field(
                                                                                String,
                                                                                description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                                                is_optional=True,
                                                                            )
                                                                        }
                                                                    )
                                                                )
                                                            }
                                                        )
                                                    ),
                                                }
                                            )
                                        ),
                                        'maxChunksBeingTransferred': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'sort': Field(
                                            PermissiveDict(
                                                fields={
                                                    'bypassMergeThreshold': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        ),
                                        'spill': Field(
                                            PermissiveDict(
                                                fields={
                                                    'compress': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        ),
                                        'accurateBlockThreshold': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'registration': Field(
                                            PermissiveDict(
                                                fields={
                                                    'timeout': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'maxAttempts': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                }
                                            )
                                        ),
                                        'memoryFraction': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                    }
                                )
                            ),
                            'eventLog': Field(
                                PermissiveDict(
                                    fields={
                                        'logBlockUpdates': Field(
                                            PermissiveDict(
                                                fields={
                                                    'enabled': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        ),
                                        'longForm': Field(
                                            PermissiveDict(
                                                fields={
                                                    'enabled': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        ),
                                        'compress': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'dir': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'enabled': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'overwrite': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'buffer': Field(
                                            PermissiveDict(
                                                fields={
                                                    'kb': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        ),
                                    }
                                )
                            ),
                            'ui': Field(
                                PermissiveDict(
                                    fields={
                                        'dagGraph': Field(
                                            PermissiveDict(
                                                fields={
                                                    'retainedRootRDDs': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        ),
                                        'enabled': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'killEnabled': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'liveUpdate': Field(
                                            PermissiveDict(
                                                fields={
                                                    'period': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        ),
                                        'port': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'retainedJobs': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'retainedStages': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'retainedTasks': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'reverseProxy': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'reverseProxyUrl': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'showConsoleProgress': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'retainedDeadExecutors': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'filters': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                    }
                                )
                            ),
                            'worker': Field(
                                PermissiveDict(
                                    fields={
                                        'ui': Field(
                                            PermissiveDict(
                                                fields={
                                                    'retainedExecutors': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'retainedDrivers': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                }
                                            )
                                        )
                                    }
                                )
                            ),
                            'sql': Field(
                                PermissiveDict(
                                    fields={
                                        'ui': Field(
                                            PermissiveDict(
                                                fields={
                                                    'retainedExecutions': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        )
                                    }
                                )
                            ),
                            'streaming': Field(
                                PermissiveDict(
                                    fields={
                                        'ui': Field(
                                            PermissiveDict(
                                                fields={
                                                    'retainedBatches': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        ),
                                        'backpressure': Field(
                                            PermissiveDict(
                                                fields={
                                                    'enabled': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'initialRate': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                }
                                            )
                                        ),
                                        'blockInterval': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'receiver': Field(
                                            PermissiveDict(
                                                fields={
                                                    'maxRate': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'writeAheadLog': Field(
                                                        PermissiveDict(
                                                            fields={
                                                                'enable': Field(
                                                                    String,
                                                                    description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                                    is_optional=True,
                                                                ),
                                                                'closeFileAfterWrite': Field(
                                                                    String,
                                                                    description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                                    is_optional=True,
                                                                ),
                                                            }
                                                        )
                                                    ),
                                                }
                                            )
                                        ),
                                        'unpersist': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'stopGracefullyOnShutdown': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'kafka': Field(
                                            PermissiveDict(
                                                fields={
                                                    'maxRatePerPartition': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'minRatePerPartition': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'maxRetries': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                }
                                            )
                                        ),
                                        'driver': Field(
                                            PermissiveDict(
                                                fields={
                                                    'writeAheadLog': Field(
                                                        PermissiveDict(
                                                            fields={
                                                                'closeFileAfterWrite': Field(
                                                                    String,
                                                                    description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                                    is_optional=True,
                                                                )
                                                            }
                                                        )
                                                    )
                                                }
                                            )
                                        ),
                                    }
                                )
                            ),
                            'broadcast': Field(
                                PermissiveDict(
                                    fields={
                                        'compress': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'blockSize': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'checksum': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                    }
                                )
                            ),
                            'io': Field(
                                PermissiveDict(
                                    fields={
                                        'compression': Field(
                                            PermissiveDict(
                                                fields={
                                                    'codec': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'lz4': Field(
                                                        PermissiveDict(
                                                            fields={
                                                                'blockSize': Field(
                                                                    String,
                                                                    description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                                    is_optional=True,
                                                                )
                                                            }
                                                        )
                                                    ),
                                                    'snappy': Field(
                                                        PermissiveDict(
                                                            fields={
                                                                'blockSize': Field(
                                                                    String,
                                                                    description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                                    is_optional=True,
                                                                )
                                                            }
                                                        )
                                                    ),
                                                    'zstd': Field(
                                                        PermissiveDict(
                                                            fields={
                                                                'level': Field(
                                                                    String,
                                                                    description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                                    is_optional=True,
                                                                ),
                                                                'bufferSize': Field(
                                                                    String,
                                                                    description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                                    is_optional=True,
                                                                ),
                                                            }
                                                        )
                                                    ),
                                                }
                                            )
                                        )
                                    }
                                )
                            ),
                            'kryo': Field(
                                PermissiveDict(
                                    fields={
                                        'classesToRegister': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'referenceTracking': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'registrationRequired': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'registrator': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'unsafe': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                    }
                                )
                            ),
                            'kryoserializer': Field(
                                PermissiveDict(
                                    fields={
                                        'buffer': Field(
                                            PermissiveDict(
                                                fields={
                                                    'root': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'max': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                }
                                            )
                                        )
                                    }
                                )
                            ),
                            'rdd': Field(
                                PermissiveDict(
                                    fields={
                                        'compress': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        )
                                    }
                                )
                            ),
                            'serializer': Field(
                                PermissiveDict(
                                    fields={
                                        'root': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'objectStreamReset': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                    }
                                )
                            ),
                            'memory': Field(
                                PermissiveDict(
                                    fields={
                                        'fraction': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'storageFraction': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'offHeap': Field(
                                            PermissiveDict(
                                                fields={
                                                    'enabled': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'size': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                }
                                            )
                                        ),
                                        'useLegacyMode': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                    }
                                )
                            ),
                            'storage': Field(
                                PermissiveDict(
                                    fields={
                                        'memoryFraction': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'unrollFraction': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'replication': Field(
                                            PermissiveDict(
                                                fields={
                                                    'proactive': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        ),
                                        'memoryMapThreshold': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                    }
                                )
                            ),
                            'cleaner': Field(
                                PermissiveDict(
                                    fields={
                                        'periodicGC': Field(
                                            PermissiveDict(
                                                fields={
                                                    'interval': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        ),
                                        'referenceTracking': Field(
                                            PermissiveDict(
                                                fields={
                                                    'root': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'blocking': Field(
                                                        PermissiveDict(
                                                            fields={
                                                                'root': Field(
                                                                    String,
                                                                    description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                                    is_optional=True,
                                                                ),
                                                                'shuffle': Field(
                                                                    String,
                                                                    description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                                    is_optional=True,
                                                                ),
                                                            }
                                                        )
                                                    ),
                                                    'cleanCheckpoints': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                }
                                            )
                                        ),
                                    }
                                )
                            ),
                            'default': Field(
                                PermissiveDict(
                                    fields={
                                        'parallelism': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        )
                                    }
                                )
                            ),
                            'hadoop': Field(
                                PermissiveDict(
                                    fields={
                                        'cloneConf': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'validateOutputSpecs': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'mapreduce': Field(
                                            PermissiveDict(
                                                fields={
                                                    'fileoutputcommitter': Field(
                                                        PermissiveDict(
                                                            fields={
                                                                'algorithm': Field(
                                                                    PermissiveDict(
                                                                        fields={
                                                                            'version': Field(
                                                                                String,
                                                                                description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                                                is_optional=True,
                                                                            )
                                                                        }
                                                                    )
                                                                )
                                                            }
                                                        )
                                                    )
                                                }
                                            )
                                        ),
                                    }
                                )
                            ),
                            'rpc': Field(
                                PermissiveDict(
                                    fields={
                                        'message': Field(
                                            PermissiveDict(
                                                fields={
                                                    'maxSize': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        ),
                                        'numRetries': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'retry': Field(
                                            PermissiveDict(
                                                fields={
                                                    'wait': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        ),
                                        'askTimeout': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'lookupTimeout': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                    }
                                )
                            ),
                            'blockManager': Field(
                                PermissiveDict(
                                    fields={
                                        'port': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        )
                                    }
                                )
                            ),
                            'network': Field(
                                PermissiveDict(
                                    fields={
                                        'timeout': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        )
                                    }
                                )
                            ),
                            'port': Field(
                                PermissiveDict(
                                    fields={
                                        'maxRetries': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        )
                                    }
                                )
                            ),
                            'core': Field(
                                PermissiveDict(
                                    fields={
                                        'connection': Field(
                                            PermissiveDict(
                                                fields={
                                                    'ack': Field(
                                                        PermissiveDict(
                                                            fields={
                                                                'wait': Field(
                                                                    PermissiveDict(
                                                                        fields={
                                                                            'timeout': Field(
                                                                                String,
                                                                                description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                                                is_optional=True,
                                                                            )
                                                                        }
                                                                    )
                                                                )
                                                            }
                                                        )
                                                    )
                                                }
                                            )
                                        )
                                    }
                                )
                            ),
                            'cores': Field(
                                PermissiveDict(
                                    fields={
                                        'max': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        )
                                    }
                                )
                            ),
                            'locality': Field(
                                PermissiveDict(
                                    fields={
                                        'wait': Field(
                                            PermissiveDict(
                                                fields={
                                                    'root': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'node': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'process': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'rack': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                }
                                            )
                                        )
                                    }
                                )
                            ),
                            'scheduler': Field(
                                PermissiveDict(
                                    fields={
                                        'maxRegisteredResourcesWaitingTime': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'minRegisteredResourcesRatio': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'mode': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'revive': Field(
                                            PermissiveDict(
                                                fields={
                                                    'interval': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        ),
                                        'listenerbus': Field(
                                            PermissiveDict(
                                                fields={
                                                    'eventqueue': Field(
                                                        PermissiveDict(
                                                            fields={
                                                                'capacity': Field(
                                                                    String,
                                                                    description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                                    is_optional=True,
                                                                )
                                                            }
                                                        )
                                                    )
                                                }
                                            )
                                        ),
                                    }
                                )
                            ),
                            'blacklist': Field(
                                PermissiveDict(
                                    fields={
                                        'enabled': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'timeout': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'task': Field(
                                            PermissiveDict(
                                                fields={
                                                    'maxTaskAttemptsPerExecutor': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'maxTaskAttemptsPerNode': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                }
                                            )
                                        ),
                                        'stage': Field(
                                            PermissiveDict(
                                                fields={
                                                    'maxFailedTasksPerExecutor': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'maxFailedExecutorsPerNode': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                }
                                            )
                                        ),
                                        'application': Field(
                                            PermissiveDict(
                                                fields={
                                                    'maxFailedTasksPerExecutor': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'maxFailedExecutorsPerNode': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'fetchFailure': Field(
                                                        PermissiveDict(
                                                            fields={
                                                                'enabled': Field(
                                                                    String,
                                                                    description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                                    is_optional=True,
                                                                )
                                                            }
                                                        )
                                                    ),
                                                }
                                            )
                                        ),
                                        'killBlacklistedExecutors': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                    }
                                )
                            ),
                            'speculation': Field(
                                PermissiveDict(
                                    fields={
                                        'root': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'interval': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'multiplier': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'quantile': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                    }
                                )
                            ),
                            'task': Field(
                                PermissiveDict(
                                    fields={
                                        'cpus': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'maxFailures': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'reaper': Field(
                                            PermissiveDict(
                                                fields={
                                                    'enabled': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'pollingInterval': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'threadDump': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'killTimeout': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                }
                                            )
                                        ),
                                    }
                                )
                            ),
                            'stage': Field(
                                PermissiveDict(
                                    fields={
                                        'maxConsecutiveAttempts': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        )
                                    }
                                )
                            ),
                            'dynamicAllocation': Field(
                                PermissiveDict(
                                    fields={
                                        'enabled': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'executorIdleTimeout': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'cachedExecutorIdleTimeout': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'initialExecutors': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'maxExecutors': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'minExecutors': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'executorAllocationRatio': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'schedulerBacklogTimeout': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'sustainedSchedulerBacklogTimeout': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                    }
                                )
                            ),
                            'r': Field(
                                PermissiveDict(
                                    fields={
                                        'numRBackendThreads': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'command': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'driver': Field(
                                            PermissiveDict(
                                                fields={
                                                    'command': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        ),
                                        'shell': Field(
                                            PermissiveDict(
                                                fields={
                                                    'command': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        ),
                                        'backendConnectionTimeout': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'heartBeatInterval': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                    }
                                )
                            ),
                            'graphx': Field(
                                PermissiveDict(
                                    fields={
                                        'pregel': Field(
                                            PermissiveDict(
                                                fields={
                                                    'checkpointInterval': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    )
                                                }
                                            )
                                        )
                                    }
                                )
                            ),
                            'deploy': Field(
                                PermissiveDict(
                                    fields={
                                        'recoveryMode': Field(
                                            String,
                                            description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                            is_optional=True,
                                        ),
                                        'zookeeper': Field(
                                            PermissiveDict(
                                                fields={
                                                    'url': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                    'dir': Field(
                                                        String,
                                                        description='''Deploy: When `spark.deploy.recoveryMode` is set to ZOOKEEPER, this configuration is used to set the zookeeper directory to store recovery state.''',
                                                        is_optional=True,
                                                    ),
                                                }
                                            )
                                        ),
                                    }
                                )
                            ),
                        }
                    )
                )
            }
        )
    )


# pylint: enable=line-too-long
